#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__all__ = ['KillCoro', 'kill_coro_on', 'try_kill_coro_on', 'akill_coro_on', 'atry_kill_coro_on', 'kill_coro', 'try_kill_coro', 'akill_coro', 'atry_kill_coro']


from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.parallel_execution.coroutines.coro_standard_services_internal_lib.service_with_a_direct_request import *
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import PutCoro
from cengal.code_flow_control.smart_values import ValueExistence
from cengal.introspection.inspect import get_exception, get_exception_tripple
from typing import Any, Optional, Tuple, Dict, Set, Union, List


class KillCoro(TypedService[bool], ServiceWithADirectRequestMixin):
    def __init__(self, loop: CoroSchedulerType):
        super(KillCoro, self).__init__(loop)
        self.direct_requests: List[Tuple[CoroID, bool]] = list()
    
    def single_task_registration_or_immediate_processing(
            self, coro_id: CoroID, tree: bool = False) -> Tuple[bool, bool, Optional[BaseException]]:
        exception = None
        result = None
        try:
            if tree:
                put_coro: PutCoro = self._loop.get_service_instance(PutCoro)
                children: Set[CoroID] = put_coro.get_set_of_all_children(coro_id)

            result = self._loop.kill_coro_by_id(coro_id)
            if tree:
                for child_coro_id in children:
                    self._loop.kill_coro_by_id(child_coro_id)
        except:
            exception = get_exception()
        
        return True, result, exception

    def full_processing_iteration(self):
        direct_requests_buff = self.direct_requests
        self.direct_requests = type(direct_requests_buff)()
        put_coro: PutCoro = self._loop.get_service_instance(PutCoro)
        for coro_id, tree in direct_requests_buff:
            try:
                if tree:
                    children: Set[CoroID] = put_coro.get_set_of_all_children(coro_id)
                
                self._loop.kill_coro_by_id(coro_id)
                if tree:
                    for child_coro_id in children:
                        self._loop.kill_coro_by_id(child_coro_id)
            except:
                ex_type, exception, tracback = get_exception_tripple()
                if __debug__: dlog(ex_type, exception, tracback)
                raise

        self.make_dead()
    
    def _add_direct_request(self, coro_id: CoroID, tree: bool = False) -> ValueExistence[None]:
        self.direct_requests.append((coro_id, tree))
        self.make_live()
        return (False, None)

    def in_work(self) -> bool:
        result = bool(self.direct_requests)
        return self.thrifty_in_work(result)


def kill_coro_on(context: Tuple[Optional[CoroSchedulerType], Optional[Interface], bool], coro_id: CoroID, tree: bool = False) -> ValueExistence[bool]:
    return make_request_to_service_with_context(context, KillCoro, coro_id, tree)


def try_kill_coro_on(context: Tuple[Optional[CoroSchedulerType], Optional[Interface], bool], coro_id: CoroID, tree: bool = False) -> ValueExistence[Optional[bool]]:
    return try_make_request_to_service_with_context(context, KillCoro, coro_id, tree)


async def akill_coro_on(context: Tuple[Optional[CoroSchedulerType], Optional[Interface], bool], coro_id: CoroID, tree: bool = False) -> ValueExistence[bool]:
    return await amake_request_to_service_with_context(context, KillCoro, coro_id, tree)


async def atry_kill_coro_on(context: Tuple[Optional[CoroSchedulerType], Optional[Interface], bool], coro_id: CoroID, tree: bool = False) -> ValueExistence[Optional[bool]]:
    return await atry_make_request_to_service_with_context(context, KillCoro, coro_id, tree)


def kill_coro(coro_id: CoroID, tree: bool = False) -> ValueExistence[bool]:
    return make_request_to_service(KillCoro, coro_id, tree)


def try_kill_coro(coro_id: CoroID, tree: bool = False) -> ValueExistence[Optional[bool]]:
    return try_make_request_to_service(KillCoro, coro_id, tree)


async def akill_coro(coro_id: CoroID, tree: bool = False) -> ValueExistence[bool]:
    return await amake_request_to_service(KillCoro, coro_id, tree)


async def atry_kill_coro(coro_id: CoroID, tree: bool = False) -> ValueExistence[Optional[bool]]:
    return await atry_make_request_to_service(KillCoro, coro_id, tree)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__all__ = ['Text', 'BinText', 'DEFAULT_ENCODING', 'EncodingRequired', 'NotSupportedDataType', 'NotSupportedDesiredTextType', 
           'normalize_text', 'normalize_text_to_data', 'find_text', 'replace_slice', 'replace_text', 'normalize_line_separators', 
           'normalize_line_separators_and_tabs', 'removeprefix', 'removesuffix', 'to_identifier', 'remove_repetitive']


#!/usr/bin/env python
# coding=utf-8


from cengal.system import PYTHON_VERSION_INT
from typing import Optional, Tuple, Union, Type, Callable, Set, List
import string
import keyword


Text = Union[bytes, bytearray, str]
BinText = Union[bytes, bytearray]
DEFAULT_ENCODING = 'utf-8'


class EncodingRequired(Exception):
    pass


class NotSupportedDesiredTextType(Exception):
    pass


class NotSupportedDataType(Exception):
    pass


def _default_normalizer(text: Text, desired_type: Type, encoding: Optional[str] = DEFAULT_ENCODING, normalizer: Optional[Callable] = None) -> Text:
    raise NotImplementedError


def normalize_text(text: Text, desired_type: Type, encoding: Optional[str] = DEFAULT_ENCODING, normalizer: Optional[Callable] = None) -> Text:
    normalizer = normalizer or _default_normalizer
    
    need_to_use_normalizer = False
    if issubclass(desired_type, bytes):
        if isinstance(text, bytes):
            pass
        elif isinstance(text, bytearray):
            text = bytes(text)
        elif isinstance(text, str):
            if encoding:
                text = text.encode(encoding)
            else:
                raise EncodingRequired
        else:
            need_to_use_normalizer = True
    elif issubclass(desired_type, bytearray):
        if isinstance(text, bytearray):
            pass
        elif isinstance(text, bytes):
            text = bytearray(text)
        elif isinstance(text, str):
            if encoding:
                text = bytearray(text, encoding)
            else:
                raise EncodingRequired
        else:
            need_to_use_normalizer = True
    elif issubclass(desired_type, str):
        if isinstance(text, str):
            pass
        elif isinstance(text, bytes) or isinstance(text, bytearray):
            if encoding:
                text = text.decode(encoding)
            else:
                raise EncodingRequired
        else:
            need_to_use_normalizer = True
    else:
        need_to_use_normalizer = True
    
    if need_to_use_normalizer:
        text = normalizer(text, desired_type, encoding)

    return text


def normalize_text_to_data(data: Text, text: Text, encoding: Optional[str] = DEFAULT_ENCODING, normalizer: Optional[Callable] = None) -> Text:
    data_type = type(data)
    if not isinstance(text, data_type):
        text = normalize_text(text, data_type, encoding, normalizer)
    
    return text


def find_text(data: Text, text: Text, start: int = 0, stop: int = -1, encoding: Optional[str] = DEFAULT_ENCODING, normalizer: Optional[Callable] = None) -> Optional[slice]:
    text = normalize_text_to_data(data, text, encoding, normalizer)
    start = data.find(text, start, stop)
    if -1 == start:
        return None
    
    return slice(start, start + len(text))


def replace_slice(data: Text, place: slice, text: Text, encoding: Optional[str] = DEFAULT_ENCODING, normalizer: Optional[Callable] = None) -> Tuple[Text, slice]:
    text = normalize_text_to_data(data, text, encoding, normalizer)
    l_text = data[:place.start]
    r_text = data[place.stop:]
    result_text = l_text + text + r_text
    result_place = slice(place.start, place.start + len(text))
    return result_text, result_place


def replace_text(data: Text, old_text: Text, new_text: Text, count: int = -1, encoding: Optional[str] = DEFAULT_ENCODING, normalizer: Optional[Callable] = None) -> Optional[Text]:
    old_text = normalize_text_to_data(data, old_text, encoding, normalizer)
    new_text = normalize_text_to_data(data, new_text, encoding, normalizer)
    return data.replace(old_text, new_text, count)


def normalize_line_separators(text: Text, encoding: Optional[str] = DEFAULT_ENCODING, normalizer: Optional[Callable] = None) -> Text:
    lines = text.splitlines()
    line_separator = '\n'
    return normalize_text(line_separator, type(text), encoding, normalizer).join(lines)


def normalize_line_separators_and_tabs(text: Text, tabsize=4, encoding: Optional[str] = DEFAULT_ENCODING, normalizer: Optional[Callable] = None) -> Text:
    text = normalize_line_separators(text, encoding, normalizer)
    return text.expandtabs(tabsize)


def removeprefix(data: Text, prefix: Text, encoding: Optional[str] = DEFAULT_ENCODING, normalizer: Optional[Callable] = None) -> Text:
    prefix = normalize_text_to_data(data, prefix, encoding, normalizer)
    if (3, 9) <= PYTHON_VERSION_INT:
        return data.removeprefix(prefix)
    else:
        if data.startswith(prefix):
            return data[len(prefix):]
        else:
            return data
        

def removesuffix(data: Text, suffix: Text, encoding: Optional[str] = DEFAULT_ENCODING, normalizer: Optional[Callable] = None) -> Text:
    suffix = normalize_text_to_data(data, suffix, encoding, normalizer)
    if (3, 9) <= PYTHON_VERSION_INT:
        return data.removesuffix(suffix)
    else:
        if data.endswith(suffix):
            return data[:-len(suffix):]
        else:
            return data
        

def to_identifier(text: Text, need_to_remove_repetitive: bool = True, encoding: Optional[str] = DEFAULT_ENCODING, normalizer: Optional[Callable] = None) -> Text:
    original_text = text
    text = normalize_text_to_data(str(), text, encoding, normalizer)
    valid_initial_chars = string.ascii_letters + '_'
    valid_chars = valid_initial_chars + string.digits
    text_chars: Set[str] = set(text)

    trans = str.maketrans({
        char: '_' for char in text_chars if char not in valid_chars
    })
    identifier = text.translate(trans)
    if need_to_remove_repetitive:
        identifier = remove_repetitive(identifier, '_')
    
    if not identifier or identifier[0] not in valid_initial_chars:
        identifier = '_' + identifier

    while keyword.iskeyword(identifier):
        identifier += '_'

    return normalize_text_to_data(original_text, text, encoding, normalizer)
        

def remove_repetitive(data: Text, sub_str: Text, encoding: Optional[str] = DEFAULT_ENCODING, normalizer: Optional[Callable] = None) -> Text:
    normalized_data: str = normalize_text_to_data(str(), data, encoding, normalizer)
    normalized_sub_str: str = normalize_text_to_data(str(), sub_str, encoding, normalizer)
    normalized_data.strip(normalized_sub_str)
    split_normalized_data: List[str] = normalized_data.split(normalized_sub_str)
    result: str = normalized_sub_str.join((piece for piece in split_normalized_data if piece))
    return normalize_text_to_data(data, result, encoding, normalizer)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


import sys, os, boto3, json, pickle, logging
from enum import Enum
from typing import Dict, Optional, Any, Tuple
from cengal.file_system.path_manager import path_relative_to_cwd
from .aws_translate__supported_languages import codes, names, name_by_code
_boto3_client = None

def aws_client():
    global _boto3_client
    if _boto3_client is None:
        aws_init()
    return _boto3_client


def aws_init(region_name: Optional[str]=None):
    global _boto3_client
    region_name = region_name or 'us-east-1'
    _boto3_client = boto3.client(service_name='translate', region_name=region_name, use_ssl=True)


_logger = None

def init_logger():
    global _logger
    _logger = logging.getLogger('text_dictionary_translator')


def log() -> logging.Logger:
    if _logger is None:
        init_logger()
    return _logger


class AwsClientForTranslateText:

    def __init__(self, cache_file_path: Optional[str]=None) -> None:
        self._cache_file_path = cache_file_path
        self._cache = None

    def _prepare(self):
        if self._cache is None:
            if self._cache_file_path is None:
                return False
            if os.path.exists(self._cache_file_path) and os.path.isfile(self._cache_file_path):
                with open(self._cache_file_path, 'rb') as file:
                    self._cache = pickle.load(file)
                    return True
            else:
                self._cache = dict()
                return True
        else:
            return True

    def save(self):
        if self._cache_file_path is not None:
            if self._cache is None:
                text_dictionary = dict()
            else:
                text_dictionary = self._cache
            with open(self._cache_file_path, 'wb') as file:
                pickle.dump(text_dictionary, file)

    def _cache_get(self, key, default: Optional[Any]=None) -> Any:
        if self._cache is not None:
            return self._cache.get(key, default)
        return default

    def _cache_set(self, key, value: Any) -> Any:
        if self._cache is not None:
            self._cache[key] = value

    def __call__(self, text, source_language_code, target_language_code):
        result = None
        if self._prepare():
            key = (
             text, source_language_code, target_language_code)
            result = self._cache_get(key)
        if result is None:
            result = aws_client().translate_text(Text=text, SourceLanguageCode=source_language_code,
              TargetLanguageCode=target_language_code)
            result = result.get('TranslatedText')
            if self._prepare():
                self._cache_set(key, result)
        return result


aws_translate: AwsClientForTranslateText = None

class TextCaseTypes(Enum):
    mixed = 0
    lower = 1
    capitalized = 2
    title = 3
    upper = 4


def detect_text_case(text: str) -> TextCaseTypes:
    if text.islower():
        return TextCaseTypes.lower
    if text.istitle():
        return TextCaseTypes.title
    if text.isupper():
        return TextCaseTypes.upper
    if text:
        if text[0].istitle():
            second_part = text[1:]
            if second_part:
                if text[1:].islower():
                    return TextCaseTypes.capitalized
            else:
                return TextCaseTypes.capitalized
    return TextCaseTypes.mixed


def apply_text_case(text: str, case: TextCaseTypes) -> str:
    if TextCaseTypes.mixed == case:
        return text
    if TextCaseTypes.lower == case:
        return text.lower()
    if TextCaseTypes.capitalized == case:
        return text.capitalize()
    if TextCaseTypes.title == case:
        return text.title()
    if TextCaseTypes.upper == case:
        return text.upper()
    raise RuntimeError(f"Unknown text case type: {case}")


def translate(text, source_language_code, target_language_code):
    case = detect_text_case(text)
    translation = aws_translate(text, source_language_code, target_language_code)
    return apply_text_case(translation, case)


def read_text_dictionary(file_path: str) -> Dict:
    with open(file_path, 'rb') as file:
        return json.load(file)


def save_text_dictionary(file_path: str, text_dictionary: Dict):
    with open(file_path, 'w', encoding='utf-8') as file:
        return json.dump(text_dictionary, file, ensure_ascii=False, indent=4)


def translate_text_dictionary(text_dictionary: Dict) -> Dict:
    translated_items = 0
    translated_variants = 0
    translation_requests = 0
    successfull_translation_requests = 0
    sent_characters = 0
    received_characters = 0
    language_in_code = text_dictionary['original_text_language']
    default_language = text_dictionary['default_language']
    translation_language_map = text_dictionary['translation_language_map']
    if default_language in translation_language_map:
        default_language = translation_language_map[default_language]
    manually_translated_to = set(text_dictionary.get('manually_translated_to', list()))
    all_translated_codes = set(codes())
    translated_codes = set()
    for code in codes():
        code = code.lower()
        if code in translation_language_map:
            code = translation_language_map[code]
        else:
            translation_language_map[code] = code
        translated_codes.add(code)

    translated_codes = list(translated_codes)
    translated_codes.sort()
    translated_to = all_translated_codes | manually_translated_to
    auto_translated_to = all_translated_codes - manually_translated_to
    auto_translated_to = list(auto_translated_to)
    auto_translated_to.sort()
    text_dictionary['auto_translated_to'] = auto_translated_to
    translated_to = list(translated_to)
    translated_to.sort()
    text_dictionary['translated_to'] = translated_to
    manually_translated_to = list(manually_translated_to)
    manually_translated_to.sort()
    text_dictionary['manually_translated_to'] = manually_translated_to
    rtl_languages = [
     'ar', 'he', 'fa', 'ur']
    rtl_languages.sort()
    text_dictionary['rtl_languages'] = rtl_languages
    present_laguages_names = text_dictionary.get('laguages_names', dict())
    laguages_english_names = dict()
    laguages_names = dict()
    log().info('Translating language names ->')
    for code in codes():
        original_name = name_by_code(code)
        laguages_english_names[code] = original_name
        if code in present_laguages_names:
            continue
        try:
            translation_requests += 1
            sent_characters += len(original_name)
            translation_result = translate(original_name, 'en', code)
            successfull_translation_requests += 1
            received_characters += len(translation_result)
            log().info(f"\t[{code}]: {translation_result}")
            laguages_names[code] = translation_result
        except:
            log().exception('\t -> [{code}] Translation Failed.')

    text_dictionary['laguages_english_names'] = laguages_english_names
    present_laguages_names.update(laguages_names)
    text_dictionary['laguages_names'] = present_laguages_names
    text_translation_list = text_dictionary['text_translation_list']
    for item in text_translation_list:
        translated_items += 1
        text_in_code = item['text']
        if 'translations' not in item:
            item['translations'] = {'default': dict()}
        else:
            translations = item['translations']
            copy_settings = item.get('copy_settings', dict())
            for tr_variant_name, tr_variant_data in translations.items():
                if tr_variant_name in copy_settings:
                    variant_copy_settings = copy_settings[tr_variant_name]
                    variant_copy_default_language = variant_copy_settings['default_language']
                    variant_copy_translation_language_map = variant_copy_settings['translation_language_map']
                    log().info('Copy translation:')
                    for code in translated_codes:
                        if code in tr_variant_data:
                            continue
                        else:
                            copy_source_language_code = variant_copy_translation_language_map.get(code, variant_copy_default_language)
                            copy_source_text = tr_variant_data[copy_source_language_code]
                            log().info(f"\t[{copy_source_language_code}] -> [{code}]: {copy_source_text}")
                            tr_variant_data[code] = copy_source_text

                    continue
                else:
                    if default_language in tr_variant_data:
                        source_language_code = default_language
                        source_text = tr_variant_data[default_language]
                    else:
                        source_language_code = language_in_code
                        source_text = text_in_code
                    log().info(f'Translating [{source_language_code}] "{source_text}" ->')
                    translated_variants += 1
                    for code in translated_codes:
                        if code == source_language_code:
                            continue
                        if code in tr_variant_data:
                            continue
                        try:
                            translation_requests += 1
                            sent_characters += len(source_text)
                            translation_result = translate(source_text, source_language_code, code)
                            successfull_translation_requests += 1
                            received_characters += len(translation_result)
                            log().info(f"\t[{code}]: {translation_result}")
                            tr_variant_data[code] = translation_result
                        except:
                            log().exception('\t -> [{code}] Translation Failed.')

    log().info('Translation Results:')
    log().info(f"\ttranslated_items: {translated_items}")
    log().info(f"\ttranslated_variants: {translated_variants}")
    log().info(f"\ttranslation_requests: {translation_requests}")
    log().info(f"\tsuccessfull_translation_requests: {successfull_translation_requests}")
    log().info(f"\tsent_characters: {sent_characters}")
    log().info(f"\treceived_characters: {received_characters}")
    return text_dictionary


def is_str_equal(text1: str, text2: str):
    return text1.casefold() == text2.casefold()


if __name__ == '__main__':
    aws_init('eu-central-1')
    logging.basicConfig()
    log().setLevel(logging.INFO)
    if 3 <= len(sys.argv):
        text_dictionary_source_file_path = sys.argv[1]
        text_dictionary_result_file_path = sys.argv[2]
        aws_translate_cache_file_path = None
        if 4 <= len(sys.argv):
            aws_translate_cache_file_path = sys.argv[3]
    else:
        text_dictionary_source_file_path = 'text_stylizer/text_dictionaly_src.json'
        text_dictionary_result_file_path = 'text_stylizer/text_dictionaly.json'
        aws_translate_cache_file_path = 'text_stylizer/text_dictionaly_cache.pickle'
    if not os.path.exists(text_dictionary_source_file_path):
        text_dictionary_source_file_path = path_relative_to_cwd(text_dictionary_source_file_path)
        if not os.path.exists(text_dictionary_source_file_path) or os.path.exists(text_dictionary_result_file_path):
            text_dictionary_result_file_path = path_relative_to_cwd(text_dictionary_result_file_path)
            if os.path.exists(text_dictionary_result_file_path):
                pass
            else:
                print(f"ERROR: Result text dictionary file not found: {text_dictionary_result_file_path}")
                text_dictionary_result_file_path = None
        else:
            print(f"ERROR: Source text dictionary file not found: {text_dictionary_source_file_path}")
            text_dictionary_source_file_path = None
    if text_dictionary_source_file_path is not None:
        if text_dictionary_result_file_path is not None:
            print('Loading text dictionary...')
            text_dictionary = read_text_dictionary(text_dictionary_source_file_path)
            print('Translating text dictionary...')
            aws_translate = AwsClientForTranslateText(aws_translate_cache_file_path)
            text_dictionary = translate_text_dictionary(text_dictionary)
            aws_translate.save()
            save_text_dictionary(text_dictionary_result_file_path, text_dictionary)
            print('Done')


#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, Set, Tuple, Dict
from math import modf
from enum import Enum
from copy import copy
from cengal.data_manipulation.conversion.mapping import inverse_mapping
from cengal.text_processing.optional_formatter.versions.v_2 import OptionalFormatter, IT, AK
from ....constants import *

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


SECONDS_PER_YEAR = SECONDS_PER_DAY * 365
SECONDS_PER_MONTH = round(SECONDS_PER_YEAR / 12)
SECONDS_PER_DECADE = SECONDS_PER_YEAR * 10
SECONDS_PER_CENTURY = SECONDS_PER_YEAR * 100
SECONDS_PER_MILLENIA = SECONDS_PER_YEAR * 1000


class TimeAttributes(Enum):
    microseconds = 0
    milliseconds = 1
    seconds = 2
    minutes = 3
    hours = 4
    days = 5
    weeks = 6
    months = 7
    years = 8
    decades = 9
    centuries = 10
    millennia = 11


ATTRIBUTES_ASCENDING = sorted(TimeAttributes, key=lambda item: item.value, reverse=False)
ATTRIBUTES_DESCENDING = sorted(TimeAttributes, key=lambda item: item.value, reverse=True)


DIVIDER_PER_ATTRIBUTE = {
    TimeAttributes.millennia: SECONDS_PER_MILLENIA,
    TimeAttributes.centuries: SECONDS_PER_CENTURY,
    TimeAttributes.decades: SECONDS_PER_DECADE,
    TimeAttributes.years: SECONDS_PER_YEAR,
    TimeAttributes.months: SECONDS_PER_MONTH,
    TimeAttributes.weeks: SECONDS_PER_WEEK,
    TimeAttributes.days: SECONDS_PER_DAY,
    TimeAttributes.hours: SECONDS_PER_HOUR,
    TimeAttributes.minutes: SECONDS_PER_MINUTE
}

ATTRIBUTE_PER_DIVIDER = inverse_mapping(DIVIDER_PER_ATTRIBUTE)

DIV_ATTRIBUTES = set(DIVIDER_PER_ATTRIBUTE)
DIV_ATTRIBUTES_ASCENDING = sorted(DIV_ATTRIBUTES, key=lambda item: item.value, reverse=False)
DIV_ATTRIBUTES_DESCENDING = sorted(DIV_ATTRIBUTES, key=lambda item: item.value, reverse=True)


MULTIPLIER_PER_ATTRIBUTE = {
    TimeAttributes.milliseconds: 1 / SECONDS_PER_MILLISECOND,
    TimeAttributes.microseconds: 1 / SECONDS_PER_MICROSECOND
}

ATTRIBUTE_PER_MULTIPLIER = inverse_mapping(MULTIPLIER_PER_ATTRIBUTE)

MUL_ATTRIBUTES = set(MULTIPLIER_PER_ATTRIBUTE)
MUL_ATTRIBUTES_ASCENDING = sorted(MUL_ATTRIBUTES, key=lambda item: item.value, reverse=False)
MUL_ATTRIBUTES_DESCENDING = sorted(MUL_ATTRIBUTES, key=lambda item: item.value, reverse=True)


DEFAULT_ATTRIBUTES = {
    TimeAttributes.years
    , TimeAttributes.months
    , TimeAttributes.days
    , TimeAttributes.hours
    , TimeAttributes.minutes
    , TimeAttributes.seconds
    , TimeAttributes.microseconds
}

DEFAULT_ATTRIBUTES_ASCENDING = sorted(DEFAULT_ATTRIBUTES, key=lambda item: item.value, reverse=False)
DEFAULT_ATTRIBUTES_DESCENDING = sorted(DEFAULT_ATTRIBUTES, key=lambda item: item.value, reverse=True)


DEFAULT_FORMATTER = OptionalFormatter(tuple(DEFAULT_ATTRIBUTES_DESCENDING), {
    TimeAttributes.years: ('',  'Y', '{}', '', '-'),
    TimeAttributes.months: ('', 'M',  '{}', '', '-'),
    TimeAttributes.days: ('', 'D', '{}', '', ' '),
    TimeAttributes.hours: ('', 'h', '{}', '', ''),
    TimeAttributes.minutes: (':', 'm', '{}', '', ''),
    TimeAttributes.seconds: (':', 's', '{}', '', ''),
    TimeAttributes.microseconds: ('.', 'μs', '{0:0>6}', '', '')
})

FULL_FORMATTER = OptionalFormatter(tuple(ATTRIBUTES_DESCENDING), {
    TimeAttributes.millennia: ('', '', 'Millennia: {}', '', ', '),
    TimeAttributes.centuries: ('', '', 'Centuries: {}', '', ', '),
    TimeAttributes.decades: ('', '', 'Decades: {}', '', ', '),
    TimeAttributes.years: ('', '', 'Years: {}', '', ', '),
    TimeAttributes.months: ('', '', 'Months: {}', '', ', '),
    TimeAttributes.weeks: ('', '', 'Weeks: {}', '', ', '),
    TimeAttributes.days: ('', '', 'Days: {}', '', ', '),
    TimeAttributes.hours: ('', '', 'Hours: {}', '', ', '),
    TimeAttributes.minutes: ('', '', 'Minutes: {}', '', ', '),
    TimeAttributes.seconds: ('', '', 'Seconds: {}', '', ', '),
    TimeAttributes.milliseconds: ('', '', 'Milliseconds: {0:0>3}', '', ', '),
    TimeAttributes.microseconds: ('', '', 'Microseconds: {0:0>6}', '', ', '),
})


class ApproximateTimeRepresentation:
    def __init__(self, seconds: Union[int, float], attributes: Union[Set[TimeAttributes], None]=None, crop: bool=False):
        self._initial__seconds = seconds
        self._initial__attributes = attributes or DEFAULT_ATTRIBUTES
        self._initial__crop = crop

        self.values = dict()

        rest = self._initial__seconds
        for attribute in DIV_ATTRIBUTES_DESCENDING:
            self.values[attribute], rest = self.compute_attribute(rest, attribute)

        rest, seconds = modf(rest)
        seconds = int(seconds)
        self.values[TimeAttributes.seconds] = seconds

        for attribute in MUL_ATTRIBUTES_DESCENDING:
            self.values[attribute], rest = self.compute_fractional_attribute(rest, attribute)

        self.cropped_values = copy(self.values)
        self._normalize()
        if self._initial__crop:
            self._crop()

    def compute_attribute(self, seconds: float,
                          attribute: TimeAttributes) -> Tuple[float, float]:
        if attribute in self._initial__attributes:
            result, rest = divmod(seconds, DIVIDER_PER_ATTRIBUTE[attribute])
            result = int(result)
            return result, rest
        else:
            return 0, seconds

    def compute_fractional_attribute(self, seconds: float,
                                     attribute: TimeAttributes) -> Tuple[float, float]:
        if attribute in self._initial__attributes:
            multiplier = MULTIPLIER_PER_ATTRIBUTE[attribute]
            rest, integer_value = modf(seconds * multiplier)
            integer_value = int(integer_value)
            rest /= multiplier
            return integer_value, rest
        else:
            return 0, seconds

    def _normalize(self):
        for attribute in TimeAttributes:
            if attribute not in self._initial__attributes:
                if attribute in self.cropped_values:
                    del self.cropped_values[attribute]

    def _crop(self):
        descending_stopped = False
        ascending_stopped = False

        for attribute in ATTRIBUTES_DESCENDING:
            descending_stopped = self._crop__check_attribute(self.values[attribute], attribute, descending_stopped)

        for attribute in ATTRIBUTES_ASCENDING:
            ascending_stopped = self._crop__check_attribute(self.values[attribute], attribute, ascending_stopped)

    def _crop__check_attribute(self, value: Union[int, float], attribute: TimeAttributes, stopped: bool) -> bool:
        if stopped or (0 != value):
            return True
        else:
            if attribute in self.cropped_values:
                del self.cropped_values[attribute]
            return False

    def format(self, formatter: Union[OptionalFormatter, None]=None) -> str:
        formatter = formatter or DEFAULT_FORMATTER
        args = tuple()
        return formatter(self._construct_formatter_dict(self.values))

    def format_cropped(self, formatter: Union[OptionalFormatter, None]=None) -> str:
        formatter = formatter or DEFAULT_FORMATTER
        return formatter(self._construct_formatter_dict(self.cropped_values))

    def __str__(self):
        formatter = FULL_FORMATTER
        return '{class_name}{{{data}}}'.format(class_name=self.__class__.__name__,
                                               data=formatter(self._construct_formatter_dict(self.values)))

    @staticmethod
    def _construct_formatter_dict(values: Dict[TimeAttributes, Union[int, float]]):
        result = dict()
        for item, value in values.items():
            result[item] = value
        return result



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cpuinfo

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


def get_cpu_info()->dict:
    return cpuinfo.get_cpu_info()


def get_l2_cache_size()->int:
    size_text = cpuinfo.get_cpu_info()['l2_cache_size']
    size_text_list = size_text.split()
    size_text_list_size = len(size_text_list)
    size_text_number = None
    size_text_dimension = None
    if 0 == size_text_list_size:
        return 0
    elif 1 == size_text_list_size:
        return int(size_text_list)
    elif 2 == size_text_list_size:
        size_text_number, size_text_dimension = size_text_list
    else:
        return 0
    size_text_number = int(size_text_number)
    size_text_dimension = size_text_dimension.lower()
    factor = 1
    if 'kb' == size_text_dimension:
        factor = 1024
    elif 'mb' == size_text_dimension:
        factor = 1024**2
    elif 'gb' == size_text_dimension:
        factor = 1024**3  # :)
    return size_text_number * factor


def l2_cache_per_core()->int:
    core_count = cpuinfo.get_cpu_info()['count']
    if core_count:
        return int(get_l2_cache_size() / core_count)
    else:
        return 0



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['ProcessDpiAwareness', 'set_dpi_awareness']


import os
from enum import IntEnum


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


"""
https://stackoverflow.com/questions/41315873/attempting-to-resolve-blurred-tkinter-text-scaling-on-windows-10-high-dpi-disp
https://stackoverflow.com/questions/44398075/can-dpi-scaling-be-enabled-disabled-programmatically-on-a-per-session-basis

typedef enum _PROCESS_DPI_AWARENESS { 
    PROCESS_DPI_UNAWARE = 0,
    /*  DPI unaware. This app does not scale for DPI changes and is
        always assumed to have a scale factor of 100% (96 DPI). It
        will be automatically scaled by the system on any other DPI
        setting. */

    PROCESS_SYSTEM_DPI_AWARE = 1,
    /*  System DPI aware. This app does not scale for DPI changes.
        It will query for the DPI once and use that value for the
        lifetime of the app. If the DPI changes, the app will not
        adjust to the new DPI value. It will be automatically scaled
        up or down by the system when the DPI changes from the system
        value. */

    PROCESS_PER_MONITOR_DPI_AWARE = 2
    /*  Per monitor DPI aware. This app checks for the DPI when it is
        created and adjusts the scale factor whenever the DPI changes.
        These applications are not automatically scaled by the system. */
} PROCESS_DPI_AWARENESS;
"""


class ProcessDpiAwareness(IntEnum):
    process_dpi_unaware = 0
    process_system_dpi_aware = 1
    process_per_monitor_dpi_aware = 2


def set_dpi_awareness(process_dpi_awareness: ProcessDpiAwareness):
    if 'nt' != os.name:
        return
    
    from ctypes import c_int, windll, byref
    from platform import win32_ver

    # Query DPI Awareness (Windows 10 and 8)
    awareness = c_int()
    errorCode = windll.shcore.GetProcessDpiAwareness(0, byref(awareness))
    # print(awareness.value)

    windows_major_version, _, _, _ = win32_ver()
    if windows_major_version in {'8', '10'}:
        # Set DPI Awareness  (Windows 10 and 8)
        errorCode = windll.shcore.SetProcessDpiAwareness(int(process_dpi_awareness))
        # the argument is the awareness level, which can be 0, 1 or 2:
        # for 1-to-1 pixel control I seem to need it to be non-zero (I'm using level 2)
    else:
        # Set DPI Awareness  (Windows 7 and Vista)
        success = windll.user32.SetProcessDPIAware()
        # behaviour on later OSes is undefined, although when I run it on my Windows 10 machine, it seems to work with effects identical to SetProcessDpiAwareness(1)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__all__ = ['PortStatus', 'Table', 'Protocol', 'purify_ports', 'get_tables', 'used_ports', 'PortLord', 'PortInfo', 'UsedPorts', 'PortsIterator', 'unify_ports', 'UnaceptablePortsRangeTypeError']

import pickle
from enum import Enum
from typing import Dict, Optional, Set, List, Tuple, Union, Any
from collections import namedtuple
from cengal.file_system.path_manager import path_relative_to_src


class PortStatus(str, Enum):
    na = 'N/A'                 # In programming APIs (not in communication between hosts), requests a system-allocated (dynamic) port
    yes = 'Yes'                # Described protocol is assigned for this port by IANA and is standardized, specified, or widely used for such.
    unofficial = 'Unofficial'  # Described protocol is not assigned for this port by IANA but is standardized, specified, or widely used for such.
    assigned = 'Assigned'      # Described protocol is assigned by IANA for this port, but is not standardized, specified, or widely used for such.
    no = 'No'                  # Described protocol is not assigned by IANA, standardized, specified, or widely used for the port.
    reserved = 'Reserved'      # Port is reserved by IANA, generally to prevent collision having its previous use removed. The port number may be available for assignment upon request to IANA.


class Table(str, Enum):
    system = 'system'
    user = 'user'
    ephemeral = 'ephemeral'
    

class Protocol(Enum):
    tcp = 0
    udp = 1
    sctp = 2
    dccp = 3


_tables: Dict[Table, List[List]] = None
_known_tables_names = {'system', 'user', 'ephemeral'}
_used_ports = None


def _purify_item(item: Optional[str]) -> PortStatus:
    if item is None:
        return PortStatus.no
    
    return PortStatus(item)


def get_tables() -> Dict[Table, List[List]]:
    global _tables
    if _tables is None:
        tables: Dict[Table, List[List]] = dict()
        for table_name in _known_tables_names:
            with open(path_relative_to_src(f'data/table_{table_name}.pickle'), 'rb') as file:
                tables[Table(table_name)] = pickle.load(file)

        _tables = tables
    
    return _tables


def used_ports() -> 'UsedPorts':
    global _used_ports
    if _used_ports is None:
        _used_ports = UsedPorts()
    
    return _used_ports


PortLord = namedtuple('PortLord', 'tcp udp sctp dccp description')


class PortInfo:
    def __init__(self, port: int, port_lords: Optional[Set[PortLord]] = None) -> None:
        self.port: int = port
        self.port_lords = port_lords or set()
    
    def add(self, port_lord: PortLord) -> int:
        self.port_lords.add(port_lord)
        return len(self.port_lords)


class UsedPorts:
    def __init__(self) -> None:
        self.table: List[PortInfo] = list([PortInfo(i) for i in range(65536)])
        self.tables = {
            Table.system: slice(0, 1024),
            Table.user: slice(1024, 49152),
            Table.ephemeral: slice(49152, 65536),
        }
        raw_tables: Dict[Table, List[List]] = get_tables()
        for table_name, raw_table in raw_tables.items():
            for row in raw_table:
                port_or_range, tcp, udp, sctp, dccp, description = row
                port_lord = PortLord(_purify_item(tcp), _purify_item(udp), _purify_item(sctp), _purify_item(dccp), description)
                if isinstance(port_or_range, tuple):
                    for port in range(int(port_or_range[0]), int(port_or_range[1]) + 1):
                        self.table[port].add(port_lord)
                else:
                    self.table[int(port_or_range)].add(port_lord)
        
        not_used_port = PortLord(PortStatus.no, PortStatus.no, PortStatus.no, PortStatus.no, str())
        for port_info in self.table:
            if not port_info.port_lords:
                port_info.port_lords.add(not_used_port)
    
    def port(self, protocol: Protocol, port_or_range: Union[int, slice, Tuple[int, int]], statuses: Union[PortStatus, Set[PortStatus]]) -> 'PortsIterator':
        return PortsIterator(self, protocol, port_or_range, statuses, 1)
    
    def range(self, protocol: Protocol, port_or_range: Union[int, slice, Tuple[int, int]], statuses: Union[PortStatus, Set[PortStatus]], desired_number_of_ports: int) -> 'PortsIterator':
        return PortsIterator(self, protocol, port_or_range, statuses, desired_number_of_ports)
    
    def __call__(self, protocol: Protocol, port_or_range: Union[int, slice, Tuple[int, int]], statuses: Union[PortStatus, Set[PortStatus]], desired_number_of_ports: int) -> 'PortsIterator':
        return self.range(protocol, port_or_range, statuses, desired_number_of_ports)


class UnaceptablePortsRangeTypeError(Exception):
    pass


class PortsIterator:
    def __init__(self, used_ports: UsedPorts, protocol: Protocol, port_or_range: Union[int, slice, Tuple[int, int]], statuses: Union[PortStatus, Set[PortStatus]], desired_number_of_ports: int) -> None:
        self.used_ports: UsedPorts = used_ports
        self.protocol: Protocol = protocol
        self.ports_range: slice = None
        if isinstance(port_or_range, int):
            self.ports_range = slice(port_or_range, port_or_range + 1)
        elif isinstance(port_or_range, slice):
            self.ports_range = port_or_range
        elif isinstance(port_or_range, tuple):
            self.ports_range = slice(*port_or_range)
        else:
            raise UnaceptablePortsRangeTypeError
        
        if isinstance(statuses, PortStatus):
            statuses = {statuses}
        
        self.statuses: Union[PortStatus, Set[PortStatus]] = statuses
        self.desired_number_of_ports: int = desired_number_of_ports
        self.first_port = self.ports_range.start

    def __iter__(self):
        self.first_port = self.ports_range.start
        return self

    def __next__(self):
        ports_range = self._iter_impl()
        self.first_port += 1
        return ports_range
    
    def _iter_impl(self) -> slice:
        if self.first_port >= self.ports_range.stop:
            raise StopIteration
        
        while self.first_port < self.ports_range.stop:
            result_range_stop = self.first_port + 1
            for offset in range(self.desired_number_of_ports):
                port = self.first_port + offset
                
                port_info: PortInfo = self.used_ports.table[port]
                port_statuses: Set[PortStatus] = set()
                for port_lord in port_info.port_lords:
                    port_statuses.add(port_lord[self.protocol.value])
                
                if 0 == len(port_statuses - self.statuses):
                    is_good = True
                else:
                    is_good = False
                
                if is_good:
                    result_range_stop = port + 1
                else:
                    result_range_stop = None
                    self.first_port = port + 1
                    break
            
            if result_range_stop is not None:
                return slice(self.first_port, result_range_stop)
        
        raise StopIteration
    
    def __call__(self) -> Optional[slice]:
        try:
            return self._iter_impl()
        except StopIteration:
            return None

    def shift_only(self, num: int = 1) -> None:
        self.first_port += num

    def shift(self, num: int = 1) -> Optional[slice]:
        self.first_port += num
        return self()

    def put_only(self, num: int) -> None:
        self.first_port = num

    def put(self, num: int) -> Optional[slice]:
        self.first_port = num
        return self()


def purify_ports(ports_range: Optional[slice]) -> Optional[Union[int, Tuple[int, int]]]:
    start = ports_range.start
    stop = ports_range.stop - 1
    if start == stop:
        return start
    else:
        return start, stop


def unify_ports(ports_range: Optional[slice]) -> Optional[Tuple[int, int]]:
    start = ports_range.start
    stop = ports_range.stop - 1
    return start, stop
    



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__all__ = ['PathDoesNotExist', 'tree_to_list', 'travers_tree', 'Tree']

#!/usr/bin/env python
# coding=utf-8



from typing import Iterable, Dict, Callable, Hashable
from collections import deque






class PathDoesNotExist(Exception):
    pass


def tree_to_list(tree: Dict)->Iterable[Hashable]:
    """
    Returns all items from a tree in a iterable container.
        tree = Tree()
        ...
        all_items =  tree_to_list(tree.get_subtree())
    Is an implementation of the Tree.get_all_children() method.
    :param tree: "dict of dict of ...". F
    :return: some fast Iterable container. deque() usually
    """
    result = deque()
    for key, subtree in tree.items():
        result.append(key)
        if subtree:
            result.extend(tree_to_list(subtree))
    return result


def travers_tree(tree: Dict, functor: Callable):
    """
    Will call 'functor' for an each item in a tree
    :param tree: current traversed subtree
    :param functor: any callable with one parameter: functor(key: Hashable)
    """
    for key, subtree in tree.items():
        functor(key)
        if subtree:
            travers_tree(subtree, functor)


def travers_tree_with_path(tree: Dict, functor: Callable, path: Iterable[Hashable]=None):
    """
    The same as travers_tree(). But also gives current path to a 'functor'
    :param tree: current traversed subtree
    :param functor: any callable with two parameters: functor(key: Hashable, path: Iterable[Hashable])
    :param path: initial actual path for the current traversed subtree
    """
    current_path = path or deque()
    for key, subtree in tree.items():
        functor(key, current_path + deque((key,)))
        if subtree:
            travers_tree_with_path(subtree, functor, current_path + deque((key,)))


class Tree:
    def __init__(self):
        self._data = dict()

    def put(self, path: Iterable[Hashable]):
        current_node = self._data
        for item in path:
            if item not in current_node:
                current_node[item] = dict()
            current_node = current_node[item]

    def get_subtree(self, path: Iterable[Hashable]=None)->Dict:
        path = path or deque()
        current_path = deque()
        current_node = self._data
        for item in path:
            current_path.append(item)
            if item in current_node:
                current_node = current_node[item]
            else:
                raise PathDoesNotExist(current_path)
        return current_node

    def get_immediate_children(self, path: Iterable[Hashable]=None)->Iterable[Hashable]:
        """
        Will return immediate children list of the given path: without children of children
        :param path:
        :return:
        """
        return deque(self.get_subtree(path))

    def get_all_children(self, path: Iterable[Hashable]=None)->Iterable[Hashable]:
        """
        Will rerun all children and children of children for the given path
        The same as tree_to_list() function but for current object and for the given path
        :param path: path to subtree
        :return:
        """
        return tree_to_list(self.get_subtree(path))

    def travers_subtree(self, functor: Callable, path: Iterable[Hashable]=None):
        """
        The same as travers_tree() function but for current object and for the given path
        :param path:
        :param functor:
        """
        travers_tree(self.get_subtree(path), functor)

    def travers_tree_with_path(self, functor: Callable, path: Iterable[Hashable]=None):
        """
        The same as travers_tree_with_path() function but for current object and for the given path
        :param functor:
        :param path:
        """
        travers_tree_with_path(self.get_subtree(path), functor, path)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['ToolTip', 'ToolTipHovered']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


import tkinter as tk
from ttkbootstrap.constants import LEFT as ttkb__LEFT, SOLID as ttkb__SOLID


class ToolTip:
    def __init__(self, widget):
        self.widget = widget
        self.tip_window = None
        self.id = None
        self.x = self.y = 0

    def show_tip(self, tip_text):
        "Display text in a tooltip window"
        self.text = tip_text
        if self.tip_window or not self.text:
            return
        x, y, _, _ = self.widget.bbox("insert")
        x = x + self.widget.winfo_rootx() + 27
        y = y + self.widget.winfo_rooty() + 27
        self.tip_window = tw = tk.Toplevel(self.widget)
        tw.wm_overrideredirect(True)
        tw.wm_geometry(f"+{x}+{y}")

        label = tk.Label(tw, text=self.text, justify=ttkb__LEFT,
                      background="#ffffe0", relief=ttkb__SOLID, borderwidth=1,
                      font=("tahoma", "8", "normal"))
        label.pack(ipadx=1)

    def hide_tip(self):
        if self.tip_window:
            self.tip_window.destroy()
            self.tip_window = None


class ToolTipHovered(ToolTip):
    def __init__(self, widget, text: str):
        super().__init__(widget)
        self.text: str = text
        self.widget.bind('<Enter>', lambda e: self.show_tip(self.text))
        self.widget.bind('<Leave>', lambda e: self.hide_tip())        



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = [
    'ASM',
    'c_uint8',
    'c_uint16',
    'c_uint32',
    'c_uint64',
    'c_float',
    'c_double',
    'asm_func',
    'asm_func_guard',
    'asm_funcs_guard',
    'run_asm',
    'run_asm__c_uint8',
    'run_asm__c_uint16',
    'run_asm__c_uint32',
    'run_asm__c_uint64',
    'run_asm__c_float',
    'run_asm__c_double',
    'run_asm__no_result',
    'asm_func_declaration',
    'declare_asm_function',
    'daf',
    'compile_asm_function',
    'caf',
    'run_asm_func',
]


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.modules_management.alternative_import import alt_import
with alt_import('cpuinfo') as cpuinfo:
    if cpuinfo is None:
        CPUINFO_PRESENT: bool = False
    else:
        CPUINFO_PRESENT = True

if not CPUINFO_PRESENT:
    raise RuntimeError('Package "py-cpuinfo" is not present. Please, install it. Version 7.0.0 or higher is required.')

if not hasattr(cpuinfo, 'ASM'):
    raise RuntimeError('Package "py-cpuinfo" is not the latest version. Please, update it. Version 7.0.0 or higher is required.')

import ctypes
from collections import namedtuple
from typing import Callable, Generator, Sequence, NamedTuple


class ASM(cpuinfo.ASM):
    def __init__(self, restype = None, argtypes = None, machine_code = None):
        # cpuinfo.ASM.__init__ has mutable default parameter which is not safe. So we fixing it here.
        argtypes = argtypes if argtypes is not None else ()
        machine_code = machine_code if machine_code is not None else []
        self.destroyed: bool = True
        super().__init__(restype, argtypes, machine_code)
    
    def compile(self):
        try:
            super().compile()
        finally:
            self.destroyed = False
    
    def __call__(self, *args):
        return self.func(*args)
    
    def free(self):
        if self.destroyed:
            return
        
        try:
            super().free()
        finally:
            self.destroyed = True
    
    def __del__(self):
        self.free()


c_uint8 = ctypes.c_uint8
c_uint16 = ctypes.c_uint16
c_uint32 = ctypes.c_uint32
c_uint64 = ctypes.c_uint64
c_float = ctypes.c_float
c_double = ctypes.c_double


def asm_func(restype = None, argtypes = None, machine_code = None) -> ASM:
    asm = ASM(restype, argtypes, machine_code)
    asm.compile()
    return asm


def asm_func_guard(asm: ASM) -> Generator[Callable, None, None]:
    try:
        yield asm.func
    finally:
        asm.free()


def asm_funcs_guard(asm_functions: Sequence[ASM]) -> Generator[Sequence[Callable], None, None]:
    try:
        yield [asm.func for asm in asm_functions]
    finally:
        for asm in asm_functions:
            asm.free()


def run_asm(restype, *machine_code):
    asm = ASM(restype, (), machine_code)
    asm.compile()
    try:
        retval = asm.func()
    finally:
        asm.free()
    
    return retval


def run_asm__c_uint8(*machine_code):
    return run_asm(c_uint8, *machine_code)


def run_asm__c_uint16(*machine_code):
    return run_asm(c_uint16, *machine_code)


def run_asm__c_uint32(*machine_code):
    return run_asm(c_uint32, *machine_code)


def run_asm__c_uint64(*machine_code):
    return run_asm(c_uint64, *machine_code)


def run_asm__c_float(*machine_code):
    return run_asm(c_float, *machine_code)


def run_asm__c_double(*machine_code):
    return run_asm(c_double, *machine_code)


def run_asm__no_result(*machine_code):
    return run_asm(None, *machine_code)


asm_func_declaration: NamedTuple = namedtuple('asm_func_declaration', ['restype', 'argtypes', 'machine_code'])  


def declare_asm_function(restype, argtypes, *machine_code) -> NamedTuple:
    return asm_func_declaration(restype, argtypes, machine_code)


daf = declare_asm_function


def compile_asm_function(declaration: NamedTuple) -> ASM:
    asm = ASM(declaration.restype, declaration.argtypes, declaration.machine_code)
    asm.compile()
    return asm


caf = compile_asm_function


def run_asm_func(declaration: NamedTuple, *args):
    asm = ASM(declaration.restype, declaration.argtypes, declaration.machine_code)
    asm.compile()
    try:
        retval = asm.func(*args)
    finally:
        asm.free()
    
    return retval



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


import unittest
from cengal.modules_management.module_rel_path import *


class TestCaseForModuleRelPath(unittest.TestCase):
    def test_module_dir(self):
        import os
        import asyncio
        from asyncio import coroutines
        from cengal.introspection.inspect import get_module_importable_str_and_path

        mdir: str = module_dir(asyncio)
        mrpath: str = module_rel_path(asyncio, coroutines)
        fpath: str = os.path.join(mdir, mrpath)
        module_file_full_path = get_module_importable_str_and_path(coroutines)[1]
        self.assertEqual(fpath, module_file_full_path)

    def test_module_rel_path(self):
        import asyncio
        from asyncio import coroutines
        self.assertEqual(module_rel_path(asyncio, coroutines), 'coroutines.py')

    def test_module_rel_import_str(self):
        import asyncio
        from asyncio import coroutines
        self.assertEqual(module_import_str(asyncio, coroutines), 'asyncio.coroutines')

    def test_current_module_dir(self):
        import os
        from cengal.introspection.inspect import entity_owning_module_info_and_owning_path, frame

        mdir: str = current_module_dir()
        _, _, module_file_full_path, _ = entity_owning_module_info_and_owning_path(frame())
        self.assertEqual(os.path.join(mdir, os.path.basename(module_file_full_path)), module_file_full_path)

    def test_current_module_rel_path(self):
        import cengal
        self.assertEqual(current_module_rel_path(cengal), 'modules_management/module_rel_path/versions/v_0/tests/test__module_rel_path.py')

    def test_current_module_import_str(self):
        import cengal
        self.assertEqual(current_module_import_str(cengal), 'cengal.modules_management.module_rel_path.versions.v_0.tests.test__module_rel_path')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__all__ = ['StreamType', 'GateSecurityPolicy', 'StreamManagerIOCoreMemoryManagement', 'UdpStreamManager', 'UdpStreamReader', 'UdpStreamReaderProtocol', 'UdpStreamWriter']


import warnings
import asyncio
from asyncio.exceptions import IncompleteReadError, LimitOverrunError
from asyncio import streams
from asyncio.streams import StreamWriter as OriginalStreamWriter
from asyncio.streams import StreamReader as OriginalStreamReader
from asyncio.streams import StreamReaderProtocol as OriginalStreamReaderProtocol
from asyncio import events
from asyncio import proactor_events
from asyncio import selector_events
try:
    from asyncio import unix_events
except ImportError:
    pass

from asyncio import coroutines
from asyncio.tasks import sleep, Task
from asyncio.futures import Future
from copy import copy
from enum import Enum
from cengal.io.asock_io.versions.v_1.recv_buff_size_computer.recv_buff_size_computer__python import RecvBuffSizeComputer
# from cengal.io.asock_io.versions.v_1.base import IOCoreMemoryManagement
from cengal.parallel_execution.asyncio.atasks import create_task
from cengal.parallel_execution.asyncio.timed_yield import TimedYield
from cengal.hardware.info.cpu import cpu_info
# from cengal.data_containers.dynamic_list_of_pieces import DynamicListOfPiecesDequeWithLengthControl
# from cengal.data_containers.fast_fifo import FIFODequeWithLengthControl, FIFOIsEmpty
# from cengal.data_manipulation.front_triggerable_variable import FrontTriggerableVariable, FrontTriggerableVariableType
from cengal.data_containers.dynamic_list_of_pieces.versions.v_1.dynamic_list_of_pieces__python import DynamicListOfPiecesDequeWithLengthControl
from cengal.code_flow_control.smart_values.versions import ValueExistence
from typing import Sequence, Tuple, Type, Set, Optional, Union, List
from .efficient_streams_base_internal import *
from .efficient_streams_base import *
from .efficient_streams_abstract import *


class UdpStreamManager(StreamManagerAbstract):
    def __init__(self) -> None:
        self.io_memory_management: StreamManagerIOCoreMemoryManagement = StreamManagerIOCoreMemoryManagement()
        self.autonomous_writer_stop_default_timeout: Optional[Union[int, float]] = 10.0
        self.output_to_client_container_type = DynamicListOfPiecesDequeWithLengthControl
        self.input_from_client_container_type = DynamicListOfPiecesDequeWithLengthControl

    async def open_connection(self, host=None, port=None, *,
                            loop=None, limit=DEFAULT_LIMIT,
                            stream_type: StreamType = StreamType.general, stream_name: str = str(),
                            protocol_greeting: Optional[str] = None, message_size_len: Optional[int] = None,
                            max_message_size_len: Optional[int] = None,
                            **kwds):
        """A wrapper for create_connection() returning a (reader, writer) pair.

        The reader returned is a UdpStreamReader instance; the writer is a
        UdpStreamWriter instance.

        The arguments are all the usual arguments to create_connection()
        except protocol_factory; most common are positional host and port,
        with various optional keyword arguments following.

        Additional optional keyword arguments are loop (to set the event loop
        instance to use) and limit (to set the buffer limit passed to the
        UdpStreamReader).

        (If you want to customize the UdpStreamReader and/or
        UdpStreamReaderProtocol classes, just copy the code -- there's
        really nothing special here except some convenience.)
        """
        if StreamType.gate == stream_type:
            raise ValueError(f'Wrong stream_type value: client can not be a "gate".')
        
        if loop is None:
            loop = events.get_event_loop()
        else:
            warnings.warn("The loop argument is deprecated since Python 3.8, "
                        "and scheduled for removal in Python 3.10.",
                        DeprecationWarning, stacklevel=2)
        
        message_protocol_settings: MessageProtocolSettings = MessageProtocolSettings(protocol_greeting, message_size_len, max_message_size_len)
        reader = UdpStreamReader(self, message_protocol_settings, limit=limit, loop=loop)
        protocol = UdpStreamReaderProtocol(self, message_protocol_settings, reader, loop=loop)
        transport, _ = await loop.create_connection(
            lambda: protocol, host, port, **kwds)
        writer = UdpStreamWriter(transport, protocol, reader, loop)
        return reader, writer

    async def start_server(self, client_connected_cb, host=None, port=None, *,
                        loop=None, limit=DEFAULT_LIMIT,
                        stream_type: StreamType = StreamType.general, stream_name: str = str(),
                        gate_security_policy: GateSecurityPolicy = GateSecurityPolicy.disabled, policy_managed_stream_names: Optional[Set[str]] = None,
                        protocol_greeting: Optional[str] = None, message_size_len: Optional[int] = None,
                        max_message_size_len: Optional[int] = None,
                        **kwds):
        """Start a socket server, call back for each client connected.

        The first parameter, `client_connected_cb`, takes two parameters:
        client_reader, client_writer.  client_reader is a UdpStreamReader
        object, while client_writer is a UdpStreamWriter object.  This
        parameter can either be a plain callback function or a coroutine;
        if it is a coroutine, it will be automatically converted into a
        Task.

        The rest of the arguments are all the usual arguments to
        loop.create_server() except protocol_factory; most common are
        positional host and port, with various optional keyword arguments
        following.  The return value is the same as loop.create_server().

        Additional optional keyword arguments are loop (to set the event loop
        instance to use) and limit (to set the buffer limit passed to the
        UdpStreamReader).

        The return value is the same as loop.create_server(), i.e. a
        Server object which can be used to stop the service.
        """
        if loop is None:
            loop = events.get_event_loop()
        else:
            warnings.warn("The loop argument is deprecated since Python 3.8, "
                        "and scheduled for removal in Python 3.10.",
                        DeprecationWarning, stacklevel=2)

        def factory():
            message_protocol_settings: MessageProtocolSettings = MessageProtocolSettings(protocol_greeting, message_size_len, max_message_size_len)
            reader =  UdpStreamReader(self, message_protocol_settings, limit=limit, loop=loop)
            protocol = UdpStreamReaderProtocol(self, message_protocol_settings, reader, client_connected_cb,
                                            loop=loop)
            return protocol

        return await loop.create_server(factory, host, port, **kwds)
    
    async def try_establish_message_protocol_server_side(self, reader: 'UdpStreamReader', writer: 'UdpStreamWriter') -> bool:
        return True
    
    async def try_establish_message_protocol_client_side(self, reader: 'UdpStreamReader', writer: 'UdpStreamWriter') -> bool:
        return True


class Server(events.AbstractServer):

    def __init__(self, loop, sockets, protocol_factory, ssl_context, backlog,
                 ssl_handshake_timeout):
        self._loop = loop
        self._sockets = sockets
        self._active_count = 0
        self._waiters = []
        self._protocol_factory = protocol_factory
        self._backlog = backlog
        self._ssl_context = ssl_context
        self._ssl_handshake_timeout = ssl_handshake_timeout
        self._serving = False
        self._serving_forever_fut = None

    def __repr__(self):
        return f'<{self.__class__.__name__} sockets={self.sockets!r}>'
    
    def get_loop(self):
        return self._loop

    def is_serving(self):
        return self._serving

    @property
    def sockets(self):
        if self._sockets is None:
            return ()
        return tuple(trsock.TransportSocket(s) for s in self._sockets)

    def close(self):
        sockets = self._sockets
        if sockets is None:
            return
        self._sockets = None

        for sock in sockets:
            self._loop._stop_serving(sock)

        self._serving = False

        if (self._serving_forever_fut is not None and
                not self._serving_forever_fut.done()):
            self._serving_forever_fut.cancel()
            self._serving_forever_fut = None

        if self._active_count == 0:
            self._wakeup()

    async def start_serving(self):
        self._start_serving()
        # Skip one loop iteration so that all 'loop.add_reader'
        # go through.
        await tasks.sleep(0, loop=self._loop)

    async def serve_forever(self):
        if self._serving_forever_fut is not None:
            raise RuntimeError(
                f'server {self!r} is already being awaited on serve_forever()')
        if self._sockets is None:
            raise RuntimeError(f'server {self!r} is closed')

        self._start_serving()
        self._serving_forever_fut = self._loop.create_future()

        try:
            await self._serving_forever_fut
        except exceptions.CancelledError:
            try:
                self.close()
                await self.wait_closed()
            finally:
                raise
        finally:
            self._serving_forever_fut = None

    async def wait_closed(self):
        if self._sockets is None or self._waiters is None:
            return
        waiter = self._loop.create_future()
        self._waiters.append(waiter)
        await waiter


class UdpStreamReader(OriginalStreamReader, StreamReaderAbstract):
    def __init__(self, manager: UdpStreamManager, message_protocol_settings: MessageProtocolSettings, *args, **kwargs) -> None:
        self._stream_manager = manager
        self._message_protocol_settings: MessageProtocolSettings = message_protocol_settings
        super().__init__(*args, **kwargs)
        self._smart_buffer: DynamicListOfPiecesDequeWithLengthControl = self._stream_manager.input_from_client_container_type(
            external_data_length=self._stream_manager.io_memory_management.global_in__data_full_size)
        self.recv_buff_size_computer = RecvBuffSizeComputer()
        cpu_info_inst = cpu_info()
        # self.recv_buff_size_computer.max_recv_buff_size = cpu_info_inst.l3_cache_size
        # self.recv_buff_size_computer.max_recv_buff_size = cpu_info_inst.l2_cache_size_per_virtual_core
        # self.recv_buff_size_computer.max_recv_buff_size = cpu_info_inst.l3_cache_size_per_virtual_core
        # self.recv_buff_size_computer.max_recv_buff_size = 3145728
        self.recv_buff_size_computer.max_recv_buff_size = 10 * 1024**2
        # self.recv_buff_size_computer.max_recv_buff_size = 1024
        # print(f'max_recv_buff_size: {self.recv_buff_size_computer.max_recv_buff_size}')
        self.limit_by_limit: bool = True
        self.limit_by_global_in__data_size_limit: bool = True
    
    async def read_max(self):
        return await self.read(self._limit)
    
    async def read_nearly_max(self):
        return await self.read_nearly(self._limit)
    
    async def read_with_counter(self):
        if self._exception is not None:
            raise self._exception

        # This used to just loop creating a new waiter hoping to
        # collect everything in self._buffer, but that would
        # deadlock if the subprocess sends more than self.limit
        # bytes.  So just call self.read(self._limit) until EOF.
        blocks = []
        counter = 0
        while True:
            block = await self.read_max()
            counter += 1
            if not block:
                break
            blocks.append(block)
        return b''.join(blocks), counter

    def __repr__(self):
        info = ['UdpStreamReader']
        if self._smart_buffer.size():
            info.append(f'{self._smart_buffer.size()} bytes')
        if self._eof:
            info.append('eof')
        if self._limit != DEFAULT_LIMIT:
            info.append(f'limit={self._limit}')
        if self._waiter:
            info.append(f'waiter={self._waiter!r}')
        if self._exception:
            info.append(f'exception={self._exception!r}')
        if self._transport:
            info.append(f'transport={self._transport!r}')
        if self._paused:
            info.append('paused')
        return '<{}>'.format(' '.join(info))

    def _maybe_resume_transport(self):
        if isinstance(self._transport, (
            proactor_events._ProactorDatagramTransport,
            selector_events._SelectorTransport,
            unix_events._UnixReadPipeTransport
            )):
            # if hasattr(self._transport, 'max_size'):
            try:
                self._transport.max_size = self.recv_buff_size_computer.recv_buff_size
                # print(f'max_size: {self._transport.max_size}')
            except AttributeError:
                pass
        else:
            print(f'Unsupported transport: {type(self._transport)}')
        
        if self._paused \
            and (
                ((not self.limit_by_limit) and (not self.limit_by_global_in__data_size_limit)) \
                or (self.limit_by_limit and (not self._limit)) \
                or (self.limit_by_limit and (self._smart_buffer.size() <= self._limit)) \
                or (self.limit_by_global_in__data_size_limit and (not self._stream_manager.io_memory_management.global_in__data_size_limit)) \
                or (self.limit_by_global_in__data_size_limit and (self._stream_manager.io_memory_management.global_in__data_full_size.value <= self._stream_manager.io_memory_management.global_in__data_size_limit.value))
            ):
            self._paused = False
            self._transport.resume_reading()

    def at_eof(self):
        """Return True if the buffer is empty and 'feed_eof' was called."""
        return self._eof and not self._smart_buffer.size()

    def feed_data(self, data):
        assert not self._eof, 'feed_data after feed_eof'

        if not data:
            return

        data_len = len(data)
        self.recv_buff_size_computer.calc_new_recv_buff_size(data_len)
        self._smart_buffer.add_piece_of_data(data)
        self._wakeup_waiter()

        if (self._transport is not None and
                not self._paused 
                and (
                    (self.limit_by_limit and (self._smart_buffer.size() > 2 * self._limit)
                    or (self.limit_by_global_in__data_size_limit and (self._stream_manager.io_memory_management.global_in__data_full_size.value > self._stream_manager.io_memory_management.global_in__data_size_limit.value)))
                )):
            try:
                self._transport.pause_reading()
            except NotImplementedError:
                # The transport can't be paused.
                # We'll just have to buffer all data.
                # Forget the transport so we don't keep trying.
                self._transport = None
            else:
                self._paused = True

    async def readline(self):
        """Read chunk of data from the stream until newline (b'\n') is found.

        On success, return chunk that ends with newline. If only partial
        line can be read due to EOF, return incomplete line without
        terminating newline. When EOF was reached while no bytes read, empty
        bytes object is returned.

        If limit is reached, ValueError will be raised. In that case, if
        newline was found, complete line including newline will be removed
        from internal buffer. Else, internal buffer will be cleared. Limit is
        compared against part of the line without newline.

        If stream was paused, this function will automatically resume it if
        needed.
        """
        sep = b'\n'
        seplen = len(sep)
        try:
            line = await self.readuntil(sep)
        except IncompleteReadError as e:
            return e.partial
        except LimitOverrunError as e:
            if self._smart_buffer.startswith(sep, e.consumed):
                self._smart_buffer.get_data(e.consumed + seplen)
            else:
                self._smart_buffer.clear()
            
            self._maybe_resume_transport()
            raise ValueError(e.args[0])
        return line

    async def readuntil(self, separator=b'\n'):
        """Read data from the stream until ``separator`` is found.

        On success, the data and separator will be removed from the
        internal buffer (consumed). Returned data will include the
        separator at the end.

        Configured stream limit is used to check result. Limit sets the
        maximal length of data that can be returned, not counting the
        separator.

        If an EOF occurs and the complete separator is still not found,
        an IncompleteReadError exception will be raised, and the internal
        buffer will be reset.  The IncompleteReadError.partial attribute
        may contain the separator partially.

        If the data cannot be read because of over limit, a
        LimitOverrunError exception  will be raised, and the data
        will be left in the internal buffer, so it can be read again.
        """
        seplen = len(separator)
        if seplen == 0:
            raise ValueError('Separator should be at least one-byte string')

        if self._exception is not None:
            raise self._exception

        # Consume whole buffer except last bytes, which length is
        # one less than seplen. Let's check corner cases with
        # separator='SEPARATOR':
        # * we have received almost complete separator (without last
        #   byte). i.e buffer='some textSEPARATO'. In this case we
        #   can safely consume len(separator) - 1 bytes.
        # * last byte of buffer is first byte of separator, i.e.
        #   buffer='abcdefghijklmnopqrS'. We may safely consume
        #   everything except that last byte, but this require to
        #   analyze bytes of buffer that match partial separator.
        #   This is slow and/or require FSM. For this case our
        #   implementation is not optimal, since require rescanning
        #   of data that is known to not belong to separator. In
        #   real world, separator will not be so long to notice
        #   performance problems. Even when reading MIME-encoded
        #   messages :)

        # `offset` is the number of bytes from the beginning of the buffer
        # where there is no occurrence of `separator`.
        offset = 0

        # Loop until we find `separator` in the buffer, exceed the buffer size,
        # or an EOF has happened.
        while True:
            buflen = self._smart_buffer.size()

            # Check if we now have enough data in the buffer for `separator` to
            # fit.
            if buflen - offset >= seplen:
                isep = self._smart_buffer.find(separator, offset)

                if isep != -1:
                    # `separator` is in the buffer. `isep` will be used later
                    # to retrieve the data.
                    break

                # see upper comment for explanation.
                offset = buflen + 1 - seplen
                if offset > self._limit:
                    raise LimitOverrunError(
                        'Separator is not found, and chunk exceed the limit',
                        offset)

            # Complete message (with full separator) may be present in buffer
            # even when EOF flag is set. This may happen when the last chunk
            # adds data which makes separator be found. That's why we check for
            # EOF *ater* inspecting the buffer.
            if self._eof:
                chunk = self._smart_buffer.get_data(self._smart_buffer.size())
                raise IncompleteReadError(chunk, None)

            # _wait_for_data() will resume reading if stream was paused.
            await self._wait_for_data('readuntil')

        if isep > self._limit:
            raise LimitOverrunError(
                'Separator is found, but chunk is longer than limit', isep)

        chunk = self._smart_buffer.get_data(isep + seplen)
        self._maybe_resume_transport()
        return bytes(chunk)

    async def read(self, n=-1):
        """Read up to `n` bytes from the stream.

        If n is not provided, or set to -1, read until EOF and return all read
        bytes. If the EOF was received and the internal buffer is empty, return
        an empty bytes object.

        If n is zero, return empty bytes object immediately.

        If n is positive, this function try to read `n` bytes, and may return
        less or equal bytes than requested, but at least one byte. If EOF was
        received before any byte is read, this function returns empty byte
        object.

        Returned value is not limited with limit, configured at stream
        creation.

        If stream was paused, this function will automatically resume it if
        needed.
        """

        if self._exception is not None:
            raise self._exception

        if n == 0:
            return b''

        if n < 0:
            # This used to just loop creating a new waiter hoping to
            # collect everything in self._buffer, but that would
            # deadlock if the subprocess sends more than self.limit
            # bytes.  So just call self.read(self._limit) until EOF.
            blocks = []
            while True:
                block = await self.read_nearly(max(self._limit, self._smart_buffer.size()))
                if not block:
                    break
                blocks.append(block)
            return b''.join(blocks)

        if not self._smart_buffer.size() and not self._eof:
            await self._wait_for_data('read')

        # This will work right even if buffer is less than n bytes
        data = self._smart_buffer.get_data(min(n, self._smart_buffer.size()))

        self._maybe_resume_transport()
        return data

    async def read_nearly(self, n=-1):
        """Read up to `n` bytes from the stream.

        If n is not provided, or set to -1, read until EOF and return all read
        bytes. If the EOF was received and the internal buffer is empty, return
        an empty bytes object.

        If n is zero, return empty bytes object immediately.

        If n is positive, this function try to read `n` bytes, and may return
        less or equal bytes than requested, but at least one byte. If EOF was
        received before any byte is read, this function returns empty byte
        object.

        Returned value is not limited with limit, configured at stream
        creation.

        If stream was paused, this function will automatically resume it if
        needed.
        """

        if self._exception is not None:
            raise self._exception

        if n == 0:
            return b''

        if n < 0:
            # This used to just loop creating a new waiter hoping to
            # collect everything in self._buffer, but that would
            # deadlock if the subprocess sends more than self.limit
            # bytes.  So just call self.read(self._limit) until EOF.
            blocks = []
            while True:
                block = await self.read_nearly(max(self._limit, self._smart_buffer.size()))
                if not block:
                    break
                blocks.append(block)
            return b''.join(blocks)

        if not self._smart_buffer.size() and not self._eof:
            await self._wait_for_data('read')

        # This will work right even if buffer is less than n bytes
        data = self._smart_buffer.get_data_nearly(n)

        self._maybe_resume_transport()
        return data

    async def readexactly(self, n):
        """Read exactly `n` bytes.

        Raise an IncompleteReadError if EOF is reached before `n` bytes can be
        read. The IncompleteReadError.partial attribute of the exception will
        contain the partial read bytes.

        if n is zero, return empty bytes object.

        Returned value is not limited with limit, configured at stream
        creation.

        If stream was paused, this function will automatically resume it if
        needed.
        """
        if n < 0:
            raise ValueError('readexactly size can not be less than zero')

        if self._exception is not None:
            raise self._exception

        if n == 0:
            return b''

        while self._smart_buffer.size() < n:
            if self._eof:
                incomplete = self._smart_buffer.get_data(self._smart_buffer.size())
                raise IncompleteReadError(incomplete, n)

            await self._wait_for_data('readexactly')

        if self._smart_buffer.size() == n:
            data = self._smart_buffer.get_data(self._smart_buffer.size())
        else:
            data = self._smart_buffer.get_data(n)

        self._maybe_resume_transport()
        return data
    
    async def readonly_exactly(self, n):
        if n < 0:
            raise ValueError('readexactly size can not be less than zero')

        if self._exception is not None:
            raise self._exception

        if n == 0:
            return b''

        while self._smart_buffer.size() < n:
            if self._eof:
                incomplete = self._smart_buffer.read_data(self._smart_buffer.size())
                raise IncompleteReadError(incomplete, n)

            await self._wait_for_data('readexactly')

        if self._smart_buffer.size() == n:
            data = self._smart_buffer.read_data(self._smart_buffer.size())
        else:
            data = self._smart_buffer.read_data(n)

        self._maybe_resume_transport()
        return data
    
    async def read_message(self):
        message_len_encoded = await self.readexactly(self._message_protocol_settings._message_size_len)
        message_len = int.from_bytes(message_len_encoded, 'little')
        return await self.readexactly(message_len)
    
    def message_awailable(self) -> bool:
        message_size_len = self._message_protocol_settings._message_size_len
        if self._smart_buffer.size() < message_size_len:
            return False

        message_len_encoded = self._smart_buffer.get_data(message_size_len)
        message_len = int.from_bytes(message_len_encoded, 'little')
        if self._smart_buffer.size() < (message_size_len + message_len):
            return False
        
        return True
    
    def transport_pause_reading(self):
        try:
            self._transport.pause_reading()
        except NotImplementedError:
            # The transport can't be paused.
            # We'll just have to buffer all data.
            # Forget the transport so we don't keep trying.
            pass
        else:
            self._paused = True
    
    def transport_resume_reading(self):
        self._paused = False
        self._transport.resume_reading()


class UdpStreamReaderProtocol(OriginalStreamReaderProtocol, StreamReaderProtocolAbstract):
    def __init__(self, manager: UdpStreamManager, message_protocol_settings: MessageProtocolSettings, *args, **kwargs) -> None:
        self._stream_manager = manager
        self._message_protocol_settings: MessageProtocolSettings = message_protocol_settings
        super().__init__(*args, **kwargs)

    def connection_made(self, transport):
        if self._reject_connection:
            context = {
                'message': ('An open stream was garbage collected prior to '
                            'establishing network connection; '
                            'call "stream.close()" explicitly.')
            }
            if self._source_traceback:
                context['source_traceback'] = self._source_traceback
            self._loop.call_exception_handler(context)
            transport.abort()
            return
        
        self._transport = transport
        reader = self._stream_reader
        if reader is not None:
            reader.set_transport(transport)

        self._over_ssl = transport.get_extra_info('sslcontext') is not None
        if self._client_connected_cb is not None:
            self._stream_writer = UdpStreamWriter(transport, self,
                                               reader,
                                               self._loop)
            res = self._client_connected_cb(reader,
                                            self._stream_writer)
            if coroutines.iscoroutine(res):
                self._loop.create_task(res)
            
            self._strong_reader = None


class UdpStreamWriter(OriginalStreamWriter, StreamWriterAbstract):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._stream_manager: UdpStreamManager = self._protocol._stream_manager
        self._output_to_client: DynamicListOfPiecesDequeWithLengthControl = self._stream_manager.output_to_client_container_type(
            external_data_length=self._stream_manager.io_memory_management.global_out__data_full_size)
        self.socket_write_fixed_buffer_size: ValueExistence = self._stream_manager.io_memory_management.socket_write_fixed_buffer_size
        self._autonomous_writer_future: Task = None
        self._autonomous_writer_future_stop_requessted: bool = False
        self._autonomous_writer_drain_enough_futures: List[Future] = list()
    
    def optimized_write(self, data):
        self._output_to_client.add_piece_of_data(data)
        # self.write(data)

    def owrite(self, data):
        return self.optimized_write(data)

    async def partial_drain(self):
        another_piece_of_data = self._output_to_client.get_data(self.socket_write_fixed_buffer_size.value)
        while another_piece_of_data:
            self.write(another_piece_of_data)
            await self.drain()
            another_piece_of_data = self._output_to_client.get_data(self.socket_write_fixed_buffer_size.value)

    async def pdrain(self):
        return await self.partial_drain()

    async def full_drain(self):
        await self.pdrain()
        rest_of_the_data_size = self._output_to_client.size()
        if rest_of_the_data_size:
            another_piece_of_data = self._output_to_client.get_data(rest_of_the_data_size)
            self.write(another_piece_of_data)
            await self.drain()

    async def fdrain(self):
        return await self.full_drain()
    
    def _release_autonomous_writer_waiters(self):
        current_size = self._output_to_client.size()
        autonomous_writer_drain_enough_futures_buff = self._autonomous_writer_drain_enough_futures
        self._autonomous_writer_drain_enough_futures = type(autonomous_writer_drain_enough_futures_buff)()
        for item in autonomous_writer_drain_enough_futures_buff:
            lower_water_size, future = item
            if current_size < lower_water_size:
                if (not future.done()) and (not future.cancelled()):
                    future.set_result(None)

            if (not future.done()) and (not future.cancelled()):
                self._autonomous_writer_drain_enough_futures.append(item)
    
    async def _autonomous_writer_impl(self):
        ty = TimedYield(0)
        while not self._autonomous_writer_future_stop_requessted:
            another_piece_of_data = self._output_to_client.get_data(self.socket_write_fixed_buffer_size.value)
            self._release_autonomous_writer_waiters()
            while another_piece_of_data:
                self.write(another_piece_of_data)
                await self.drain()
                another_piece_of_data = self._output_to_client.get_data(self.socket_write_fixed_buffer_size.value)
                self._release_autonomous_writer_waiters()
            
            await ty()
    
    def start_autonomous_writer(self):
        self._autonomous_writer_future = create_task(self._autonomous_writer_impl)
    
    def start_aw(self):
        return self.start_autonomous_writer()
    
    async def stop_autonomous_writer(self, timeout: Optional[Union[int, float]] = 0):
        """_summary_

        Args:
            timeout (Optional[Union[int, float]], optional): _description_. Defaults to 0. 0 - infinit timeout; None - default timeout

        Returns:
            _type_: _description_
        """
        result = None
        if timeout is None:
            timeout = self._stream_manager.autonomous_writer_stop_default_timeout
        
        if self._autonomous_writer_future and (not self._autonomous_writer_future_stop_requessted):
            self._autonomous_writer_future_stop_requessted = True
            if timeout:
                result = await asyncio.wait_for(self._autonomous_writer_future, timeout)
            else:
                result = await self._autonomous_writer_future
            
            self._autonomous_writer_future = None
            self._autonomous_writer_future_stop_requessted = False
        
        return result
    
    async def stop_aw(self, timeout: Optional[Union[int, float]] = 0):
        """_summary_

        Args:
            timeout (Optional[Union[int, float]], optional): _description_. Defaults to 0. 0 - infinit timeout; None - default timeout

        Returns:
            _type_: _description_
        """
        return await self.stop_autonomous_writer(timeout)
    
    async def autonomous_writer_drain_enough(self, lower_water_size: Optional[int] = None):
        if lower_water_size is None:
            lower_water_size = self.socket_write_fixed_buffer_size.value * 3
            # print(f'lower_water_size: {lower_water_size}')
            # lower_water_size = cpu_info().l3_cache_size
        
        if lower_water_size <= self._output_to_client.size():
            future: Future = self._loop.create_future()
            self._autonomous_writer_drain_enough_futures.append((lower_water_size, future))
            await future
    
    async def aw_drain_enough(self):
        await self.autonomous_writer_drain_enough()
    
    def optimized_write_message(self, data):
        self.optimized_write(len(data).to_bytes(self._protocol._message_protocol_settings._message_size_len, 'little') + data)
    
    def owrite_message(self, data):
        self.optimized_write_message(data)
    
    async def send_message(self, data):
        self.optimized_write_message(data)
        await self.fdrain()
